/* Copyright 2012 Tobias Marschall
 * 
 * This file is part of CLEVER.
 * 
 * CLEVER is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CLEVER is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CLEVER.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <fstream>
#include <fstream>
#include <vector>
#include <cassert>

#include <boost/program_options.hpp>
#include <bamtools/api/BamReader.h>

#include "Histogram.h"
#include "HistogramBasedDistribution.h"
#include "GroupWiseBamReader.h"
#include "VersionInfo.h"

using namespace std;
namespace po = boost::program_options;

void usage(const char* name, const po::options_description& options_desc) {
	cerr << "Usage: " << name << " [options]" << endl;
	cerr << endl;
	cerr << "Reads BAM format from stdin and outputs histogram of insert lengths" << endl;
	cerr << "of UNIQUELY MAPPED reads. Here, insert length refers to the area between" << endl;
	cerr << "two alignments (excluding the alignments)." << endl;
	cerr << endl;
	cerr << "IMPORTANT: Assumes that, in the BAM input, alignments belonging to the same read (pair)" << endl;
	cerr << "are grouped, i.e. are given in subsequent lines." << endl;
	cerr << endl;
	cerr << options_desc << endl;
	exit(1);
}

bool is_clipped(const BamTools::BamAlignment& alignment) {
	vector<BamTools::CigarOp>::const_iterator it = alignment.CigarData.begin();
	for (;it!=alignment.CigarData.end(); ++it) {
		switch (it->Type) {
		case 'S':
		case 'H':
		case 'P':
			return true;
		default:
			continue;
		}
	}
	return false;
}

void process_read(const vector<BamTools::BamAlignment*>& alignments1, const vector<BamTools::BamAlignment*>& alignments2, Histogram& histogram) {
	if ((alignments1.size()==1) && (alignments2.size()==1) && (alignments1[0]->IsReverseStrand()!=alignments2[0]->IsReverseStrand()) && (alignments1[0]->RefID==alignments2[0]->RefID)) {
		int insert_size = 0;
		if (alignments1[0]->Position <= alignments2[0]->Position) {
			insert_size = alignments2[0]->Position - alignments1[0]->GetEndPosition();
		} else {
			insert_size = alignments1[0]->Position - alignments2[0]->GetEndPosition();
		}
		histogram.add(insert_size);
	}
}

bool is_unique(const BamTools::BamAlignment& alignment) {
	if (alignment.MapQuality == 0) return false;
	if (!alignment.IsPrimaryAlignment()) return false;
	uint32_t x0_tag = -1;
	uint32_t x1_tag = -1;
	if (alignment.GetTag("X0",x0_tag)) {
		if (x0_tag>1) return false;
	}
	if (alignment.GetTag("X1",x1_tag)) {
		if (x1_tag>0) return false;
	}
	return true;
}

int main(int argc, char* argv[]) {
	VersionInfo::checkAndPrintVersion("insert-length-histogram", cerr);
	string commandline = VersionInfo::commandline(argc, argv);

	// PARAMETERS
	int min_counts_per_bin = 20;
	string insert_length_histogram = "";
	string mean_sd_filename = "";
	bool gaussian_values = false;

	po::options_description options_desc("Allowed options");
	options_desc.add_options()
		("min_count_per_bin,c", po::value<int>(&min_counts_per_bin)->default_value(20), "Minimum number of counts per bin. If necessary, bins are joined to reach this number of counts.")
		("mean_and_sd,m", po::value<string>(&mean_sd_filename), "Write (robustly estimated) mean and standard deviation of main peak to given filename.")
		("gaussian_values,g", po::value<bool>(&gaussian_values)->zero_tokens(), "Print forth column: probability based on Gaussian estimated from main peak.")
	;
	
	if (isatty(fileno(stdin))) {
		usage(argv[0], options_desc);
	}

	po::variables_map options;
	try {
		po::store(po::parse_command_line(argc, argv, options_desc), options);
		po::notify(options);
	} catch(exception& e) {
		cerr << "Error: " << e.what() << "\n";
		return 1;
	}
	cerr << "Commandline: " << commandline << endl;

	if (min_counts_per_bin<1) {
		cerr << "Error: option -c: integer >0 expected, but got " << min_counts_per_bin << "." << endl;
	}

	GroupWiseBamReader* bam_reader = 0;
	// number of alignments used in histogram
	long long total_count = 0;
	// actual histogram data
	Histogram histogram;
	try {
		bam_reader = new GroupWiseBamReader("/dev/stdin", true);
		bam_reader->enableProgressMessages(cerr, 1000000);
		while ( bam_reader->hasNext() ) {
			bam_reader->advance();
			if (bam_reader->isFirstUnmapped() || bam_reader->isSecondUnmapped()) continue;
			const vector<BamTools::BamAlignment*>& alignments1 = bam_reader->getAlignmentsFirst();
			const vector<BamTools::BamAlignment*>& alignments2 = bam_reader->getAlignmentsSecond();
			if ((alignments1.size() != 1) || (alignments2.size() != 1)) continue;
			if (alignments1[0]->IsReverseStrand() == alignments2[0]->IsReverseStrand()) continue;
			if (alignments1[0]->RefID != alignments2[0]->RefID) continue;
			if (is_clipped(*(alignments1[0])) || is_clipped(*(alignments2[0]))) continue;
			if (!is_unique(*(alignments1[0])) || !is_unique(*(alignments2[0]))) continue;
			total_count += 1;
			process_read(alignments1, alignments2, histogram);
		}
	} catch(exception& e) {
		cerr << "Error: " << e.what() << "\n";
		return 1;
	}
	delete bam_reader;

	if (total_count < 100) {
		cerr << "Too few unique read pairs (" << total_count << "). Aborting." << endl;
		return 1;
	}
	if (histogram.size() < 10) {
		cerr << "Too few distinct insert sizes (" << histogram.size() << "). Aborting." << endl;
		return 1;
	}
	cerr << "Using " << total_count << " unique read pairs with " << histogram.size() << " distinct insert sizes." << endl;
	double mean;
	double sd;
	histogram.computeMeanAndStddev(&mean, &sd);
	cerr << "Main peak of internal segment length distribution: mean " << mean << ", sd " << sd << endl;
	if (mean_sd_filename.compare("") != 0) {
		ofstream ofs(mean_sd_filename.c_str());
		ofs << mean << " " << sd << endl;
		ofs.close();
	}
	auto_ptr<HistogramBasedDistribution> ild = histogram.toDistribution(min_counts_per_bin);
	if (gaussian_values) {
		ild->printWithGaussian(cout, mean, sd);
	} else {
		cout << *ild;
	}
	return 0;
}
